"""
Actions related to user information
"""
from argparse import ArgumentParser
from typing import Iterable, Optional

from spotify_model import (
    Artist,
    CursorPaging,
    Paging,
    PublicUser,
    SavedAlbum,
    SavedTrack,
    SimplifiedPlaylist,
)
from spotipy import Spotify

from .util import exhaust, exhaust_cursor, read_credentials_oauth


def current_user(client: Spotify) -> PublicUser:
    """
    Get all details of the current user

    Required scopes:
    - user-read-private

    Optional scopes:
    - user-read-email

    https://developer.spotify.com/documentation/web-api/reference/#endpoint-get-current-users-profile
    """
    return PublicUser(**client.current_user())


def current_user_library_albums(client: Spotify) -> Iterable[SavedAlbum]:
    """
    Get the albums a user has saved to their library

    https://developer.spotify.com/documentation/web-api/reference/#endpoint-get-users-saved-albums
    """

    def _library_albums(limit: int, offset: int) -> Paging:
        return Paging(**client.current_user_saved_albums(limit, offset))

    for album in exhaust(_library_albums):
        yield SavedAlbum(**album)


def current_user_library_tracks(client: Spotify) -> Iterable[SavedTrack]:
    """
    https://developer.spotify.com/documentation/web-api/reference/#endpoint-get-users-saved-tracks
    """

    def _library_tracks(limit: int, offset: int) -> Paging:
        return Paging(**client.current_user_saved_tracks(limit, offset))

    for track in exhaust(_library_tracks):
        yield SavedTrack(**track)


def current_user_playlists(client: Spotify) -> Iterable[SimplifiedPlaylist]:
    """
    Get all playlists that belong to the current user
    """

    def _playlists(limit: int, offset: int) -> Paging:
        return Paging(**client.current_user_playlists(limit, offset))

    for playlist in exhaust(_playlists):
        yield SimplifiedPlaylist(**playlist)


def current_user_followed_artists(client: Spotify) -> Iterable[Artist]:
    """
    Required scopes:
    - user-follow-read

    https://developer.spotify.com/documentation/web-api/reference/#endpoint-get-followed
    """

    def _followed_artists(limit: int, after: Optional[str]) -> CursorPaging:
        # This one is a little different; rather than using an offset counter,
        # we inspect the response to get the "last"
        return CursorPaging(**client.current_user_followed_artists(limit=limit, after=after)["artists"])

    for artist in exhaust_cursor(_followed_artists):
        yield Artist(**artist)


def main() -> None:
    "Simple runner for retrieving current user information"
    parser = ArgumentParser()
    parser.add_argument("-c", "--credentials", required=True)
    parser.add_argument("-r", "--redirect-uri", required=True)
    cmdline = parser.parse_args()

    client = read_credentials_oauth(
        cmdline.credentials,
        redirect_uri=cmdline.redirect_uri,
        scopes=["user-read-private", "user-follow-read"],
    )

    print("Current user:")
    print(current_user(client))

    print()
    print("Current user followed artists:")
    for artist in current_user_followed_artists(client):
        print(artist)


if __name__ == "__main__":
    main()
